#/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: extime_itifg.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

/* Feature test switches */

/* System headers */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <termios.h>

#include <sys/ioctl.h>
#include <sys/param.h>

#ifndef __NetBSD__
#include <getopt.h>
#endif

#ifdef linux
#  include <linux/limits.h>	/* PATH_MAX */
#endif

/* Local headers */

#include "itifgExt.h"
#include "iticonfig.h"

/* Macros */

/* File scope variables */

static int device_desc = -1;

static struct termios dump_tty_state;

/* External variables */

extern char *optarg;
extern int optind, opterr, optopt;

/* External functions */

/* Structures and unions */

/* Signal catching functions */

/* Functions */

static void restore_tty_state (void);

void
restore_tty_state (void)
{
  tcsetattr (0, TCSAFLUSH, &dump_tty_state);
  printf ("\n");
}

/* Main */

int
main (int argc, char *argv[])
{
  int option;
  int interactive = TRUE, camera_mode = -1;
  
  char command;
  char range = '5';	/* adjust to DEF_EXPOSURE_TIME! */
  int error, board = 0, module = ICP_AMNO, modules[ITI_BOARDS_MAX];

  unsigned long long extime = 0ULL, new_extime;
  struct timespec extime_str;

  struct termios set_tty_state;

  char device_name[PATH_MAX];

  while (TRUE)
    {
      if ((option = getopt (argc, argv, "b:e:m:?")) == -1)
	break;
      switch (option)
	{
	case 'b':
	  if (!optarg)
	    printf ("No board selected, using default (%d).\n", board);
	  else
	    {
	      board = atoi (optarg);
	      if (board < 0 || board >= ITI_BOARDS_MAX)
		{
		  printf ("Wrong board selected (%d).\n", board);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 'e':
	  if (!optarg)
	    printf ("No exposure time given, using interactive mode.\n");
	  else
	    {
	      extime = atof (optarg) * 1000.0;
	      /* for now test with the lowest/highest values */ 
	      if (extime < PCD_DLY_MIN || extime > PCD_DLY_MAX)
		{
		  printf ("Wrong exporure time given (%f).\n",
			  (double)extime / 1000.0);
		  exit (EXIT_FAILURE);
		}
	      interactive = FALSE;
	    }
	  break;
	case 'm':
	  if (!optarg)
	    printf ("No camera mode given, so don't set it.\n");
	  else
	    {
	      camera_mode = atoi (optarg);
	      /* for now test with the lowest/highest values */ 
	      if (camera_mode < 0 || camera_mode > 15)
		{
		  printf ("Wrong camera mode given (%d).\n", camera_mode);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case '?':
	default:
	  printf ("Usage: extime_itifg -b board (0) [0-%d]\n",
		  ITI_BOARDS_MAX - 1);
	  printf ("                    -m camera mode\n");
	  printf ("                       am-dig (not set) [0-7]\n");
	  printf ("                       pc-dig (not set) [0-15]\n");
	  printf ("                    -e exposure time\n");
	  printf ("                       "
		  "am-dig (read out) [%07.3fus-%07.3fms]\n",
		  (double)DG_DLY_MIN / 1000.0,
		  (double)DG_DLY_MAX / 1000000.0);
	  printf ("                       "
	          "pc-dig (read_out) [%07.3fus-%07.3fms]\n",
		  (double)PCD_DLY_MIN / 1000.0,
		  (double)PCD_DLY_MAX / 1000000.0);
	  printf ("Interactive commands:  <  - minimal value\n");
	  printf ("                       -  - decrement one step\n");
	  printf ("                      0-6 - select range\n");
	  printf ("                       +  - increment one step\n");
	  printf ("                       >  - maximal value\n");
	  printf ("                       q  - exit program\n");
	  exit (EXIT_FAILURE);
	  break;
	}
    }
  
  snprintf (device_name, PATH_MAX, "%s%d%s", ITI_NAME_PREFIX, board,
	    ITI_NAME_DATA_POSTFIX);
  
  if ((device_desc = open (device_name, O_RDONLY)) < 0)
    {
      perror ("open");
      exit (EXIT_FAILURE);
    }
  if ((error = read_procfs (NULL, NULL, NULL, modules)))
    {
      printf ("Error while reading the procfs entry.\n");
      exit (EXIT_FAILURE);
    }
  module = modules[board];

  /* now test for the higher/lower values */
  if (module == ICP_AMDG)
    {
      if (extime < DG_DLY_MIN || extime > DG_DLY_MAX)
	{
	  printf ("Wrong exporure time given (%f).\n",
		  (double)extime / 1000.0);
	  exit (EXIT_FAILURE);
	}
      if (camera_mode < 0 || camera_mode > 7)
	{
	  printf ("Wrong camera mode given (%d).\n", camera_mode);
	  exit (EXIT_FAILURE);
	}
    }

  if (camera_mode != -1)
    {
      switch (module)
	{
	case ICP_AMDG:
	  if (ioctl (device_desc, GIOC_DG_SET_CAMMD, &camera_mode) < 0)
	    {
	      perror ("GIOC_DG_SET_CAMMD");
	      exit (EXIT_FAILURE);
	    }
	  break;
	case ITI_VMPCD:
	  if (ioctl (device_desc, HIOC_PCD_SET_CAMMD, &camera_mode) < 0)
	    {
	      perror ("HIOC_PCD_SET_CAMMD");
	  exit (EXIT_FAILURE);
	    }
	  break;
	default:
	  printf ("Frame grabber board doesn't support this feature.\n");
	  exit (EXIT_FAILURE);
	}
    }

  if (interactive)
    {
      if (tcgetattr (0, &dump_tty_state) < 0)
	{
	  perror ("tcgetattr");
	  exit (EXIT_FAILURE);
	}
      if (atexit (restore_tty_state) < 0)
	{
	  perror ("atexit");
      exit (EXIT_FAILURE);
	}
      cfmakeraw (&set_tty_state);
      if (tcsetattr (0, TCSAFLUSH, &set_tty_state) < 0)
	{
	  perror ("tcsetattr");
	  exit (EXIT_FAILURE);
	}
      switch (module)
	{
	case ICP_AMDG:
	  if (ioctl (device_desc, GIOC_DG_GET_EXTIME, &extime_str) < 0)
	    {
	      perror ("GIOC_DG_GET_EXTIME");
	      exit (EXIT_FAILURE);
	    }
	  break;
	case ITI_VMPCD:
	  if (ioctl (device_desc, HIOC_PCD_GET_EXTIME, &extime_str) < 0)
	    {
	      perror ("HIOC_PCD_GET_EXTIME");
	      exit (EXIT_FAILURE);
	    }
	  break;
	default:
	  printf ("Frame grabber board doesn't support this feature.\n");
	  exit (EXIT_FAILURE);
	}
      extime = extime_str.tv_sec * 1000000000 + extime_str.tv_nsec;
      do
	{
	  printf ("<,-,0-6,+,>,q ");
	  if (extime < 1000000)
	    printf ("[%c|%07.3f us]: ", range, (float)extime / 1000.0);
	  else if (extime < 1000000000)
	    printf ("[%c|%07.3f ms]: ", range, (float)extime / 1000000.0);
	  else
	    printf ("[%c|%07.3f s]: ", range, (float)extime / 1000000000.0);
	  extime_str.tv_sec = extime / 1000000000;
	  extime_str.tv_nsec = extime % 1000000000;
	  switch (module)
	    {
	    case ICP_AMDG:
	      if (ioctl (device_desc, GIOC_DG_SET_EXTIME, &extime_str) < 0)
		{
		  perror ("GIOC_DG_SET_EXTIME");
		  exit (EXIT_FAILURE);
		}
	      break;
	    case ITI_VMPCD:
	      if (ioctl (device_desc, HIOC_PCD_SET_EXTIME, &extime_str) < 0)
		{
		  perror ("HIOC_PCD_SET_EXTIME");
		  exit (EXIT_FAILURE);
		}
	      break;
	    }
	  command = getchar();
	  switch (command)
	    {
	    case '0':
	      if (module == ITI_VMPCD)
		extime = ITI_DLY_BASE * 512, range = '0';
	      break;

	    case '1':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 1) * 1024;
	      else
		extime = (ITI_DLY_BASE << 2) * 512;
	      range = '1';
	      break;

	    case '2':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 3) * 1024;
	      else
		extime = (ITI_DLY_BASE << 4) * 512;
	      range = '2';
	      break;

	    case '3':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 5) * 1024;
	      else
		extime = (ITI_DLY_BASE << 6) * 512;
	      range = '3';
	      break;

	    case '4':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 7) * 1024;
	      else
		extime = (ITI_DLY_BASE << 8) * 512;
	      range = '4';
	      break;

	    case '5':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 9) * 1024;
	      else
		extime = (ITI_DLY_BASE << 10) * 512;
	      range = '5';
	      break;

	    case '6':
	      if (module == ICP_AMDG)
		extime = (ITI_DLY_BASE << 11) * 1024;
	      else
		extime = (ITI_DLY_BASE << 12) * 512;
	      range = '6';
	      break;

	    case '<':
	      if (module == ICP_AMDG)
		extime = DG_DLY_MIN, range = '1';
	      else
		extime = PCD_DLY_MIN, range = '0';
	      break;

	    case '>':
	      if (module == ICP_AMDG)
		extime = DG_DLY_MAX;
	      else
		extime = PCD_DLY_MAX - PCD_ROTIME_INIT * 1000.0;
	      range = '6';
	      break;

	    case '+':
	      if (extime > (ITI_DLY_BASE << 9) * 4098)
		{
		  range = '6';
		  new_extime = extime + (ITI_DLY_BASE << 11);
		}
	      else
		if (extime > (ITI_DLY_BASE << 7) * 4098)
		  {
		    range = '5';
		    if ((new_extime = extime + (ITI_DLY_BASE << 9))
			> (ITI_DLY_BASE << 9) * 4098)
		      new_extime = (ITI_DLY_BASE << 11) * 1025;
		  }
		else
		  if (extime > (ITI_DLY_BASE << 5) * 4098)
		    {
		      range = '4';
		      if ((new_extime = extime + (ITI_DLY_BASE << 7))
			  > (ITI_DLY_BASE << 7) * 4098)
			new_extime = (ITI_DLY_BASE << 9) * 1025;
		    }
		  else
		    if (extime > (ITI_DLY_BASE << 3) * 4098)
		      {
			range = '3';
			if ((new_extime = extime + (ITI_DLY_BASE << 5))
			    > (ITI_DLY_BASE << 5) * 4098)
			  new_extime = (ITI_DLY_BASE << 7) * 1025;
		      }
		    else
		      if (extime > (ITI_DLY_BASE << 1) * 4098)
			{
			  range = '2';
			  if ((new_extime = extime + (ITI_DLY_BASE << 3))
			      > (ITI_DLY_BASE << 3) * 4098)
			    new_extime = (ITI_DLY_BASE << 5) * 1025;
			}
		      else
			{
			  range = '1';
			  if ((new_extime = extime + (ITI_DLY_BASE << 3))
			      > (ITI_DLY_BASE << 1) * 4098)
			    new_extime = (ITI_DLY_BASE << 3) * 1025;
			}
	      if (module == ICP_AMDG)
		{
		  if (new_extime < DG_DLY_MAX)
		    extime = new_extime;
		}
	      else
		{
		  if (new_extime < PCD_DLY_MAX)
		    extime = new_extime;
		}
	      break;

	    case '-':
	      if (extime > (ITI_DLY_BASE << 9) * 4098)
		{
		  range = '6';
		  if ((new_extime = extime - (ITI_DLY_BASE << 11))
		      < (ITI_DLY_BASE << 11) * 1025)
		    new_extime = (ITI_DLY_BASE << 9) * 4098;
		}
	      else
		if (extime > (ITI_DLY_BASE << 7) * 4098)
		  {
		    range = '5';
		    if ((new_extime = extime - (ITI_DLY_BASE << 9))
			< (ITI_DLY_BASE << 9) * 1025)
		      new_extime = (ITI_DLY_BASE << 7) * 4098;
		  }
		else
		  if (extime > (ITI_DLY_BASE << 5) * 4098)
		    {
		      range = '4';
		      if ((new_extime = extime - (ITI_DLY_BASE << 7))
			  < (ITI_DLY_BASE << 7) * 1025)
			new_extime = (ITI_DLY_BASE << 5) * 4098;
		    }
		  else
		    if (extime > (ITI_DLY_BASE << 3) * 4098)
		      {
			range = '3';
			if ((new_extime = extime - (ITI_DLY_BASE << 5))
			    < (ITI_DLY_BASE << 5) * 1025)
			  new_extime = (ITI_DLY_BASE << 3) * 4098;
		      }
		    else
		      if (extime > (ITI_DLY_BASE << 1) * 4098)
			{
			  range = '2';
			  if ((new_extime = extime - (ITI_DLY_BASE << 3))
			      < (ITI_DLY_BASE << 3) * 1025)
			    new_extime = (ITI_DLY_BASE << 1) * 4098;
			}
		      else
			{
			  range = '1';
			  new_extime = extime - (ITI_DLY_BASE << 1);
			}
	      if (module == ICP_AMDG)
		{
		  if (new_extime > DG_DLY_MIN)
		    extime = new_extime;
		}
	      else
		{
		  if (new_extime > PCD_DLY_MIN)
		    extime = new_extime;
		}
	      break;

	    case 'q':
	      if (close (device_desc) < 0)
		{
		  perror ("close");
		  exit (EXIT_FAILURE);
		}
	      exit (EXIT_SUCCESS);
	      break;

	    default:
	      break;
	    }

	  printf ("\r");
	} while (1);
    }
  else
    {
      /* not interactive */
      extime_str.tv_sec = extime / 1000000000;
      extime_str.tv_nsec = extime % 1000000000;
      switch (module)
	{
	case ICP_AMDG:
	  if (ioctl (device_desc, GIOC_DG_SET_EXTIME, &extime_str) < 0)
	    {
	      perror ("GIOC_DG_SET_EXTIME");
	      exit (EXIT_FAILURE);
	    }
	  break;
	case ITI_VMPCD:
	  if (ioctl (device_desc, HIOC_PCD_SET_EXTIME, &extime_str) < 0)
	    {
	      perror ("HIOC_PCD_SET_EXTIME");
	      exit (EXIT_FAILURE);
	    }
	  break;
	default:
	  printf ("Frame grabber board doesn't support this feature.\n");
	  exit (EXIT_FAILURE);
	}
    }
  
  if (close (device_desc) < 0)
    {
      perror ("close");
      exit (EXIT_FAILURE);
    }

  exit (EXIT_SUCCESS);
}

