/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: test_itifg.c,v 7.0.2.1 2000/11/29 15:04:07 cvs Exp $
 *
 */

/* Feature test switches */

/* System headers */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <time.h>
#include <termios.h>

#include <math.h>

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/wait.h>

#ifndef __NetBSD__
#include <getopt.h>
#endif

#ifdef linux
#  include <linux/limits.h>	/* PATH_MAX */
#  include <asm/page.h>		/* PAGE_SIZE */
#endif

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <X11/extensions/XShm.h>

#ifdef SAVE_TIFF
  #include <tiffio.h>
#endif

/* Local headers */

#include "itifgExt.h"
#include "amvsReg.h"
#include "amdigReg.h"
#include "ampcvReg.h"
#include "pcdigReg.h"
#include "iticonfig.h"

/* Macros */

/* #define	DEBUG_ITIFG */

#define FONT_NAME  	"8x13"
#define CREATOR_PROG	"itifg"

#define EXIT_SIGNAL	2

/* File scope variables */

static int child = 0;
static pid_t children[ITI_BOARDS_MAX] = { 0, 0, 0, 0, 0, 0, 0, 0 };

static volatile int timeout = FALSE;
static volatile int record = FALSE;

static jmp_buf jmpstate;

static struct sigaction ignore;

static int devdesc = -1, rawdesc = -1, zerodesc = -1;
static caddr_t dataptr = NULL;
static size_t mapsize = 0UL;
static Display *display = NULL;
static XShmSegmentInfo shminfo;
static Window window;
static GC context = 0;
static XImage *ximage = NULL;

static struct image_t
{
  caddr_t srcptr;
  caddr_t dstptr;
  off_t srcoff, dstoff;
  size_t raw_size, paged_size;
  short width, height;
  int srcbpp, srcbpl, dstbpp, dstbpl;
  u_long lut[256];
} image =
{
  NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, {}
};

/* External variables */

extern const struct mod_names_t mod_names[];

extern char *optarg;
extern int optind, opterr, optopt;

/* External functions */

extern Status XShmQueryExtension (Display *display);

/* Structures and unions */

typedef enum
{
  READ_MODE,
  MMAP_MODE,
} opmode_t;

static struct opmode_name_t
{
  opmode_t number;
  char name[8];
} opmode_name[] =
  {
    { READ_MODE, "read" },
    { MMAP_MODE, "mmap" },
  };

typedef enum
{
  BLOCK_MODE,
  SIGNAL_MODE,
  SELECT_MODE,
  POLL_MODE
} syncmd_t;

static struct syncmd_name_t
{
  syncmd_t number;
  char name[8];
} syncmd_name[] =
  {
    { BLOCK_MODE, "block" },
    { SIGNAL_MODE, "signal" },
    { SELECT_MODE, "select" },
    { POLL_MODE, "poll" }
  };

typedef enum
{
  NOSPECIAL_MODE,
} sacqmd_t;

static struct sacqmd_name_t
{
  sacqmd_t number;
  char name[16];
} sacqmd_name[] =
  {
    { NOSPECIAL_MODE, "no special" },
  };

#define BM	ITI_BOARDS_MAX + 1  
struct args_t
{
  int conffile_i, savefile_i;
  char conffile[BM][PATH_MAX], savefile[BM][PATH_MAX];
  int save_i, display_i, wait_i, useshm_i, match_i, lock_i;
  int save[BM], display[BM], wait[BM], useshm[BM], match[BM], lock[BM];
  int  board_i, module_i, camera_i, depth_i,
    scales_i, buffer_i, frames_i, zoom_i;
  int board[BM], module[BM], camera[BM], depth[BM],
    scales[BM], buffer[BM], frames[BM], zoom[BM];
  int rate_i;
  float rate[BM];
  int opmode_i;
  opmode_t opmode[BM];
  int syncmd_i;
  syncmd_t syncmd[BM];
  int sacqmd_i;
  sacqmd_t sacqmd[BM];
  };
#undef BM

/* Signal catching functions */

void Sigio_Catcher (void);
void Sigalrm_Catcher (void);
void Sigint_Catcher (void);
void Sigchld_Catcher (void);
void Sigquit_Catcher (void);

void Clean_Up (int);

void
Sigio_Catcher (void)
{
  timeout = FALSE;

  if (sigaction (SIGIO, &ignore, NULL) < 0)
    {
      perror ("Can't install IGNORE handler");
      Clean_Up (EXIT_FAILURE);
    }
}

void
Sigalrm_Catcher (void)
{
  timeout = TRUE;

  if (sigaction (SIGIO, &ignore, NULL) < 0)
    {
      perror ("Can't install IGNORE handler");
      Clean_Up (EXIT_FAILURE);
    }
}

void
Sigint_Catcher (void)
{
  int i = 1;
  
  if (child == 0)
    while (children[i] != 0)
      kill (children[i++], SIGINT);
	
  signal (SIGINT, SIG_DFL);
  longjmp (jmpstate, 1);
}

void
Sigchld_Catcher (void)
{
  int i = 1;
  
  if (child == 0)
    while (children[i] != 0)
      kill (children[i++], SIGINT);
	
  signal (SIGINT, SIG_DFL);
  longjmp (jmpstate, 1);
}

void
Sigquit_Catcher (void)
{
  int i = 1;

  if (child == 0)
    while (children[i] != 0)
      kill (children[i++], SIGQUIT);

  record = TRUE;
}
 
/* Functions */

int Parse_Args (int, char**, struct args_t*, int*, int*);
int Init_Devs (const struct args_t*);
int Setup_Bufs (const struct args_t*);
int Startup_X11 (const struct args_t*);

int Display_Image (const struct args_t*);
int Save_Image (const struct args_t*, int, char, char*, int);

void
Clean_Up (int how)
{
  struct iti_acc_t acctinfo;

  if (devdesc != -1)
    {
      if (ioctl (devdesc, GIOC_GE_GET_STATS, &acctinfo) < 0)
	{
	  perror ("GIOC_GE_GET_STATS");
	  exit (EXIT_FAILURE);
	}
      fprintf (stderr, "\nrun statistics:\n");
      fprintf (stderr, "\tframes captured %ld,\n",
	       acctinfo.frames_captured);
      fprintf (stderr, "\tframes transfered %ld,\n",
	       acctinfo.frames_transfered);
      fprintf (stderr, "\tframes timedout %ld,\n",
	       acctinfo.frames_timedout);
      fprintf (stderr, "\tframes get_by_blocked %ld.\n",
	       acctinfo.frames_get_by_blocked);
      fprintf (stderr, "\tframes get_by_signal %ld.\n",
	       acctinfo.frames_get_by_signal);
      fprintf (stderr, "\tframes get_by_select %ld.\n",
	       acctinfo.frames_get_by_select);
    }

  if (shminfo.shmaddr)
    {
      XShmDetach (display, &shminfo);
      shmdt (shminfo.shmaddr);
      shmctl (shminfo.shmid, IPC_RMID, 0);
      ximage->data = NULL;
    }
  if (ximage != NULL)
    XDestroyImage (ximage);
  if (context != 0)
    XFreeGC (display, context);
  if (display != NULL)
    XCloseDisplay (display);

  if (dataptr != NULL)
    {
      if (mapsize == 0)
	free (dataptr);
      else
	munmap (dataptr, mapsize);
    }

  if (zerodesc != -1)
    {
      if (dataptr != NULL && mapsize != 0)
	munmap (dataptr, mapsize);
      close (zerodesc);
    }

  if (rawdesc != -1)
    close (rawdesc);

  if (devdesc != -1)
    close (devdesc);

  if (child == 0)
    {
      int i = 1, status;
      
      while (children[i] != 0)
	{
	  waitpid (children[i], &status, 0);
	  if (WIFEXITED(status))
	    fprintf (stderr, "Child %d exited with status %d.\n", i++,
		     WEXITSTATUS(status));
	}
    }

  exit (how);
}

int
Display_Image (const struct args_t *args)
{
  int line, pixel, shift;
  
  u_char *src_8 = NULL, *dst_8 = NULL;
  u_short *src_16 = NULL, *dst_16 = NULL;
  u_long *src_32 = NULL, *dst_32 = NULL;
  
  switch (image.srcbpp)
    {
    case 8:
      /*
       * 8-bit source
       */
  
      src_8 = (u_char*)(image.srcptr + image.srcoff);
      switch (image.dstbpp)
	{
	case 8:
	  /*
	   * 8-bit source, 8-bit destination
	   */
 
	  dst_8 = (u_char*)(image.dstptr + image.dstoff);
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width; pixel++, dst_8++, src_8++)
		*dst_8 = image.lut[*src_8];
	      src_8 += image.srcbpl - image.width;
	      dst_8 += image.dstbpl - image.width;
	    }
	  break;
	case 16:
	  /*
	   * 8-bit source, 16-bit destination
	   */
 
	  dst_16 = (u_short*)(image.dstptr + image.dstoff);
	  for (line = 0; line < image.height; line++) 
	    {
	      for (pixel = 0; pixel < image.width; pixel++, dst_16++, src_8++)
		*dst_16 = image.lut[*src_8];
	      src_8  += image.srcbpl - image.width;
	      dst_16 += image.dstbpl / 2 - image.width;
	    }
	  break;
	case 24:
	case 32:
	  /*
	   * 8-bit source, 24/32-bit destination
	   */
 
	  dst_32 = (u_long*)(image.dstptr + image.dstoff);
	  for(line=0; line < image.height; line++)
	    {
	      for(pixel=0; pixel < image.width; pixel++, dst_32++,src_8++)
		{
		  u_char *putC = (u_char *)dst_32;
		  putC[0] = putC[1] = putC[2] = putC[3] = *src_8;
		}
	      src_8 += image.srcbpl - image.width;
	      dst_32 += image.dstbpl / 4 - image.width;
	    }
	  break;
	default:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	}
      break;
    case 10:
    case 12:
    case 14:
    case 16:
      /*
       * 16-bit source
       */
      src_16 = (u_short*)(image.srcptr + image.srcoff);
      shift = image.srcbpp - 8;
      switch (image.dstbpp)
	{
	case 8:
	  /*
	   * 16-bit source, 8-bit destination
	   */
 
	  dst_8 = (u_char*)(image.dstptr + image.dstoff);
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width; pixel++, dst_8++, src_16++)
		*dst_8 = image.lut[*src_16 >> shift];
	      src_16 += image.srcbpl - image.width;
	      dst_8  += image.dstbpl - image.width;
	    }
	  break;
	case 16:
	  /*
	   * 16-bit source, 16-bit destination
	   */
 
	  dst_16 = (u_short*)(image.dstptr + image.dstoff);
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width; pixel++, dst_16++, src_16++)
		*dst_16 = image.lut[*src_16 >> shift];
	      src_16 += image.srcbpl / 2 - image.width;
	      dst_16 += image.dstbpl / 2 - image.width;
	    }
	  break;
	case 24:
	case 32:
	  /*
	   * 16-bit source, 24/32-bit destination
	   */
 
	  dst_32 = (u_long*)(image.srcptr + image.srcoff);
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width; pixel++, dst_32++, src_16++)
		*dst_16 = image.lut[*src_16 >> shift];
	      src_16 += image.srcbpl / 2 - image.width;
	      dst_16 += image.dstbpl / 2 - image.width;
	    }
	  break;
	default:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	}
      break;
    case 24:
      /*
       * 24-bit source, 2432-bit destination
       * 8/16-bit destination not yet supported
       */
  
      switch (image.dstbpp)
	{
	case 8:
	case 16:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	case 24:		
	case 32:
	  /*
	   * Same routine for 24/32-bit destination.
	   * This code won't work for packed 24-bit
	   * destinations.
	   * Assumption is that the most significant byte
	   * is unused in 32-bit mode.
	   */
 
	  src_8  = (u_char *)(image.srcptr + image.srcoff);
	  dst_32 = (u_long *)(image.dstptr + image.dstoff);
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width;
		   pixel++, dst_32++, src_8 += 3)
		{
		  u_long	temp;
 
		  temp = 	*src_8 
		    + (src_8[1] << 8) 
		    + (src_8[2] << 16);
		  *dst_32 = temp;
		}
	      src_8  += image.srcbpl / 3 - image.width;
	      dst_32 += image.dstbpl / 4 - image.width;
	    }
	  break;
	default:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	}
      break;
    case 32:
      /*
       * 32-bit source, 24/32-bit destination
       * 8/16-bit destination not yet supported
       */
 
      src_32 = (u_long*)(image.srcptr + image.srcoff);
 
      switch (image.dstbpp)
	{
	case 8:
	case 16:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	case 24:
	case 32:
	  /*
	   * 32-bit source, 24/32-bit destination
	   * This code won't work for packed 24-bit
	   * destinations.
	   * TRUE-COLOR MODE
	   */
 
	  src_32 = (u_long*)(image.srcptr + image.srcoff);
	  dst_32 = (u_long*)(image.dstptr + image.dstoff);
 
	  for (line = 0; line < image.height; line++)
	    {
	      for (pixel = 0; pixel < image.width;
		   pixel++, dst_32++, src_32++)
		*dst_32 = *src_32;
	      src_32 += image.srcbpl / 4 - image.width;
	      dst_32 += image.dstbpl / 4 - image.width;
	    }
	  break;
	default:
	  fprintf(stderr, "Unsupported depth combination."
		  "Source: %d bpp, destination: %d bpp\n",
		  image.srcbpp, image.dstbpp);
	  return ERROR;
	}
      break;
    default:
      fprintf(stderr, "Unsupported depth combination."
	      "Source: %d bpp, destination: %d bpp\n",
	      image.srcbpp, image.dstbpp);
      return ERROR;
    }
  
  ITI_PRINTF(".");
  return OK;
}
  
int
Save_Image (const struct args_t *args, int number, char index,
	    char *description, int create)
{
#if SAVE_TIFF
  TIFF *tiffdesc;
  int line; 
  u_char *src_8 = NULL;
  u_short *src_16 = NULL;
  u_long *src_32 = NULL;
#endif
  
  char *savename = NULL;
  char *nameptr;
  
  if (create)
    {
      if ((savename = alloca (PATH_MAX)) == NULL)
	{
	  perror ("Can't alloc temp string");
	  return ERROR;
	}
      
      nameptr = savename;
      nameptr += sprintf (nameptr, "%s", args->savefile[args->savefile_i]);
      if (args->savefile_i < child + 1)
	nameptr += sprintf (nameptr, "_%d", child);
      nameptr += sprintf (nameptr, "-%03d", number);
      nameptr += sprintf (nameptr, "%c", index + 'a');
#if SAVE_TIFF  
      nameptr += sprintf (nameptr, ".tif");
#else
      nameptr += sprintf (nameptr, ".raw");
#endif

      if ((rawdesc = open (savename, O_WRONLY | O_CREAT | O_TRUNC,
			   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0)
	{
	  fprintf (stderr, "Can't open save file (%s): %s\n", savename,
		   strerror (errno));
	  return ERROR;
	}
    }
  
#if SAVE_TIFF  
  if ((tiffdesc = TIFFFdOpen (rawdesc, savename, "wb")) == NULL)
    {
      perror ("Can't open TIFF save file");
      return ERROR;
    }

  /* specify various tiff format parameters */
  TIFFSetField (tiffdesc, TIFFTAG_IMAGEWIDTH, (unsigned long) image.width);
  TIFFSetField (tiffdesc, TIFFTAG_IMAGELENGTH, (unsigned long) image.height);

  TIFFSetField (tiffdesc, TIFFTAG_MINSAMPLEVALUE, 0L);
  TIFFSetField (tiffdesc, TIFFTAG_MAXSAMPLEVALUE, (1L << image.srcbpp));
  TIFFSetField (tiffdesc, TIFFTAG_SAMPLESPERPIXEL, (image.srcbpp + 7) / 8);
  TIFFSetField (tiffdesc, TIFFTAG_BITSPERSAMPLE, image.srcbpp);

  TIFFSetField (tiffdesc, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField (tiffdesc, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);

  TIFFSetField (tiffdesc, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
  TIFFSetField (tiffdesc, TIFFTAG_COMPRESSION, COMPRESSION_NONE); 

  TIFFSetField (tiffdesc, TIFFTAG_ROWSPERSTRIP, 1);

  TIFFSetField(tiffdesc, TIFFTAG_SOFTWARE, CREATOR_PROG);
  TIFFSetField (tiffdesc, TIFFTAG_IMAGEDESCRIPTION, description);

  switch (image.srcbpp)
    {
    case 8:
      src_8 = (u_char*)(image.srcptr + image.srcoff);
      for (line = 0; line < image.height; line++)
	{
	  if (TIFFWriteScanline(tiffdesc, src_8, line, 0) != 1)
	    {
	      TIFFClose (tiffdesc);
	      return ERROR;
	    }
	  src_8 += image.srcbpl;
	}
      break;
    case 10:
    case 12:
    case 14:
    case 16:
      src_16 = (u_short*)(image.srcptr + image.srcoff);
      for (line = 0; line < image.height; line++)
	{
	  if (TIFFWriteScanline(tiffdesc, src_16, line, 0) != 1)
	    {
	      TIFFClose (tiffdesc);
	      return ERROR;
	    }
	  src_16 += image.srcbpl / 2;
	}
      break;
    case 24:
    case 32:
      src_32 = (u_long*)(image.srcptr + image.srcoff);
      for (line = 0; line < image.height; line++)
 	{
 	  if (TIFFWriteScanline(tiffdesc, src_32, line, 0) != 1)
 	    {
 	      TIFFClose (tiffdesc);
 	      return ERROR;
 	    }
 	  src_32 += image.srcbpl / 4;
 	}
      break;
    default:
      return ERROR;
    }
  
  TIFFClose (tiffdesc);
#else
  if (write (rawdesc, image.srcptr + image.srcoff,
	     image.srcbpl * image.height) != image.srcbpl * image.height)
    {
      struct iti_acc_t acctinfo;

      if (ioctl (devdesc, GIOC_GE_GET_STATS, &acctinfo) < 0)
	{
	  perror ("GIOC_GE_GET_STATS");
	  exit (EXIT_FAILURE);
	}
      fprintf (stderr, "Can't write additional data to save file: %s",
	       strerror (errno));
      fprintf (stderr, "\nrun statistics:\n");
      fprintf (stderr, "\tframes captured %ld,\n",
	       acctinfo.frames_captured);
      fprintf (stderr, "\tframes transfered %ld,\n",
	       acctinfo.frames_transfered);
      fprintf (stderr, "\tframes timedout %ld,\n",
	       acctinfo.frames_timedout);
      fprintf (stderr, "\tframes get_by_blocked %ld.\n",
	       acctinfo.frames_get_by_blocked);
      fprintf (stderr, "\tframes get_by_signal %ld.\n",
	       acctinfo.frames_get_by_signal);
      fprintf (stderr, "\tframes get_by_select %ld.\n",
	       acctinfo.frames_get_by_select);
      Clean_Up (EXIT_FAILURE);
    }
#endif
  
  /* FIXME - fsync (rawdesc) or fdatasync (rawdesc) */
  if (!args->wait[args->wait_i])
    sync ();

    if (create)
     close (rawdesc);

  ITI_PRINTF(",");

  return OK;
}

/* ARGSUSED */
int
Parse_Args (int argc, char *argv[], struct args_t *args,
	    int *progs, int *dev_modules)
{
  int board, arg, confdesc;
  
  int error;
  u_long version;
  char date[128];
  int dev_boards;
  
  while (1)
    {
       if ((arg = getopt (argc, argv, "a:b:c:d:ef:g:iklmn:o:p:r:s:t:vz:#:?"))
 	  == -1)
  	break;
        switch (arg)
  	{
  	case 'a':
 	  if (!optarg)
 	    fprintf (stderr, "No acquisition mode selected,"
		     " using '%s' mode.\n",
		     sacqmd_name[args->sacqmd[args->sacqmd_i]].name);
 	  else
 	    {
	      args->sacqmd_i++;
 	      if (!strcmp (optarg, "none"))
 		args->sacqmd[args->sacqmd_i] = NOSPECIAL_MODE;
 	      else
		{
		  fprintf (stderr, "Wrong special acqisition mode"
			   " selected (%s).\n", optarg);
		  exit (EXIT_FAILURE);
		}
 	    }
	  break;
	case 'b':
	  if (!optarg)
	    fprintf (stderr, "No board selected, using default (%d).\n",
		    args->board[args->board_i]);
	  else
	    {
	      args->board_i++;
	      args->board[args->board_i] = atoi (optarg);
	      if (args->board[args->board_i] < 0 ||
		  args->board[args->board_i] >= ITI_BOARDS_MAX)
		{
		  fprintf (stderr, "Wrong board selected (%s).\n", optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 'c':
	  if (!optarg)
	    fprintf (stderr, "No camera selected, using default (%d).\n",
		    args->camera[args->camera_i]);
	  else
	    {
	      args->camera_i++;
	      args->camera[args->camera_i] = atoi (optarg);
	      if (args->camera[args->camera_i] < 0 ||
		  args->camera[args->camera_i] > ITI_CAMERAS_MAX)
		{
		  fprintf (stderr, "Wrong camera selected (%s).\n", optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 'd':
	  if (!optarg)
	    fprintf (stderr, "No image depth selected, using default (%d).\n",
		    args->depth[args->depth_i]);
	  else
	    {
	      args->depth_i++;
	      args->depth[args->depth_i] = atoi (optarg);
	      switch (args->depth[args->depth_i])
		{
		case 8:
		case 10:
		case 12:
		case 14:
		case 16:
		case 24:
		case 32:
		  break;
		default:
		  fprintf (stderr, "Wrong image depth selected (%s).\n",
			   optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
 	case 'e':
	  args->useshm_i++;
 	  args->useshm[args->useshm_i] = FALSE;
 	  break;
	case 'f':
	  if (!optarg)
	    fprintf (stderr, "No config file selected, using default (%s).\n",
		    args->conffile[args->conffile_i]);
	  else
	    {
	      args->conffile_i++;
	      strcpy (args->conffile[args->conffile_i], optarg);
	      if (strcspn (args->conffile[args->conffile_i], "/") ==
		  strlen (args->conffile[args->conffile_i]))
		{
		  strcpy (args->conffile[args->conffile_i], DEF_CNFDIR);
		  strcat (args->conffile[args->conffile_i], optarg);
		}
	      if ((confdesc = open (args->conffile[args->conffile_i],
				    O_RDONLY)) < 0)
		{
		  fprintf (stderr, "Open config file (%s): %s.\n", optarg,
			  strerror (errno));
		  exit (EXIT_FAILURE);
		}
	      close (confdesc);
	    }
	  break;
	case 'g':
	  if (!optarg)
	    fprintf (stderr, "No greyscales selected, using default (%d).\n",
		    args->scales[args->scales_i]);
	  else
	    {
	      args->scales_i++;
	      args->scales[args->scales_i] = atoi (optarg);
	      switch (args->scales[args->scales_i])
		{
		case 8:
		case 16:
		case 32:
		case 64:
		case 128:
		case 256:
		  break;
		default:
		  fprintf (stderr, "Wrong greyscales selected (%s).\n",
			   optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 'k':
	  args->wait_i++;
	  args->wait[args->wait_i] = TRUE;
	  break;
 	case 'i':
	  args->display_i++;
 	  args->display[args->display_i] = FALSE;
	  args->save_i++;
 	  args->save[args->save_i] = TRUE;
 	  break;
 	case 'l':
	  args->lock_i++;
 	  args->lock[args->lock_i] = TRUE;
 	  break;
 	case 'm':
	  args->match_i++;
 	  args->match[args->match_i] = TRUE;
 	  break;
	case 'n':
	  if (!optarg)
	    fprintf (stderr, "No buffer selected, using default (%d).\n",
		    args->buffer[args->buffer_i]);
	  else
	    {
	      args->buffer_i++;
	      args->buffer[args->buffer_i] = atoi (optarg);
	      if (args->buffer[args->buffer_i] < 1 ||
		  args->buffer[args->buffer_i] > ITI_SNAPS_MAX)
		{
		  fprintf (stderr, "Wrong buffer selected (%s).\n", optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 'o':
	  if (!optarg)
	    fprintf (stderr, "No operation mode selected, using '%s' mode.\n",
		    opmode_name[args->opmode[args->opmode_i]].name);
	  else
	    {
	      args->opmode_i++;
	      if (!strcmp (optarg, "read"))
		args->opmode[args->opmode_i] = READ_MODE;
	      else
		if (!strcmp (optarg, "mmap"))
		  args->opmode[args->opmode_i] = MMAP_MODE;
		else
		  {
		    fprintf (stderr, "Wrong operation mode selected (%s).\n",
			     optarg);
		    exit (EXIT_FAILURE);
		  }
	    }
	  break;
	case 'r':
	  if (!optarg)
	    fprintf (stderr, "No frames per second selected,"
		     " using default (%.2f).\n", args->rate[args->rate_i]);
	  else
	    {
	      args->rate_i++;
	      args->rate[args->rate_i] = atof (optarg);
	      if (args->rate[args->rate_i] < ITI_FPS_MIN ||
		  args->rate[args->rate_i] >= ITI_FPS_MAX)
		{
		  fprintf (stderr, "Wrong frames per second selected (%s).\n",
			  optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case 's':
	  if (!optarg)
	    fprintf (stderr, "No sync mode selected, using '%s' mode.\n",
		    syncmd_name[args->syncmd[args->syncmd_i]].name);
	  else
	    {
	      args->syncmd_i++;
	      if (!strcmp (optarg, "block"))
		args->syncmd[args->syncmd_i] = BLOCK_MODE;
	      else
		if (!strcmp (optarg, "signal"))
		  args->syncmd[args->syncmd_i] = SIGNAL_MODE;
		else
		  if (!strcmp (optarg, "select"))
		    args->syncmd[args->syncmd_i] = SELECT_MODE;
		  else
		    if (!strcmp (optarg, "poll"))
		      args->syncmd[args->syncmd_i] = POLL_MODE;
		    else
		      {
			fprintf (stderr, "Wrong sync mode selected (%s).\n",
				 optarg);
			exit (EXIT_FAILURE);
		      }
	    }
	  break;
	case 't':
 	  if (!optarg)
 	    fprintf (stderr, "No frame count selected, using default (%d).\n",
 		    args->frames[args->frames_i]);
 	  else
 	    {
	      args->frames_i++;
 	      args->frames[args->frames_i] = atoi (optarg);
 	      if (args->frames[args->frames_i] < 1)
 		{
 		  fprintf (stderr, "Wrong frame count selected (%s).\n",
			   optarg);
 		  exit (EXIT_FAILURE);
 		}
 	    }
	  break;
	case 'v':
	  args->save_i++;
	  args->save[args->save_i] = TRUE;
	  break;
	case 'z':
	  if (!optarg)
	    fprintf (stderr, "No zoom factor selected, using default (%d).\n",
		    args->zoom[args->zoom_i]);
	  else
	    {
	      args->zoom_i++;
	      args->zoom[args->zoom_i] = atoi (optarg);
	      switch (args->zoom[args->zoom_i])
		{
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
		  break;
		default:
		  fprintf (stderr, "Wrong zoom factor selected (%s).\n",
			   optarg);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case '#':
	  if (!optarg)
	    fprintf (stderr, "No number of programs selected,"
		     "using default (%d).\n", *progs);
	  else
	    {
	      *progs = atoi (optarg);
	      if (*progs < 1 || *progs > ITI_BOARDS_MAX)
		{
		  fprintf (stderr, "Wrong number of programs selected (%d).\n",
			   *progs);
		  exit (EXIT_FAILURE);
		}
	    }
	  break;
	case '?':
	default:
 	  fprintf (stderr, "Usage: %s [options] [save file (%s)]\n", argv[0],
		  args->savefile[0]);
	  fprintf (stderr, "Valid options are:\n");
	  fprintf (stderr, "\t-# use # progs   (%d) [1-4]\n", *progs);
 	  fprintf (stderr, "\t-f config File   (%s)\n", args->conffile[0]);
 	  fprintf (stderr, "\t-b Board number  (%d) [0-%d]\n", args->board[0],
		   ITI_BOARDS_MAX - 1);
 	  fprintf (stderr, "\t-c Camera number (%d) [0|1|2|3]\n",
		   args->camera[0]);
 	  fprintf (stderr, "\t-d image Depth   (%d) [8|10|12|14|16]\n",
		   args->depth[0]);
 	  fprintf (stderr, "\t-g Gray scales   (%d)[8|16|...|128|256]\n",
		  args->scales[0]);
 	  fprintf (stderr, "\t-z Zoom factor   (%d) [1|2|4|8|16]\n",
		   args->zoom[0]);
 	  fprintf (stderr, "\t-n use # buffers (%d) [1-%d]\n", args->buffer[0],
		  ITI_SNAPS_MAX);
 	  fprintf (stderr, "\t-t frame counT   (%d) [>0]\n", args->frames[0]);
 	  fprintf (stderr, "\t-r frame Rate    (%.2f)  [%.2f-%.2f]\n",
		   args->rate[0], ITI_FPS_MIN, ITI_FPS_MAX);
 	  fprintf (stderr, "\t-o Operation mode    (read)  "
		  "[read|mmap]\n");
 	  fprintf (stderr, "\t-s Synchron. mode    (block) "
		  "[block|signal|select|poll]\n");
 	  fprintf (stderr, "\t-a special Acq. mode (none)  "
		  "[none]\n");
 	  fprintf (stderr, "\t-m config file and hardware have to Match\n");
 	  fprintf (stderr, "\t-l Lock the read buffer into memory \n");
 	  fprintf (stderr, "\t-e don't use shared memory Extension\n");
 	  fprintf (stderr, "\t-i don't dIsplay, only save\n");
 	  fprintf (stderr, "\t-v also saVe, not only display\n");
 	  fprintf (stderr, "\t-k wait for a Keypress to snap\n");
	  exit (EXIT_FAILURE);
	}
    }
  for (board = 0; board < ITI_BOARDS_MAX; board++)
    if (argc > optind)
      {
	args->savefile_i++;
	strcpy (args->savefile[args->savefile_i], argv[optind++]);
      }
  while (argc > optind)
    fprintf (stderr, "Additional argument '%s' not used.\n", argv[optind++]);

  fprintf (stderr, "\n");
  if ((error = read_procfs (&version, date, &dev_boards, dev_modules)))
    {
      switch (error)
	{
	case -EITI_ARG:
	  fprintf (stderr, "read_procfs: Wrong call argument\n"
		   "(version:%p, date:%p, boards:%p, modules:%p).\n",
		   &version, date, &dev_boards, dev_modules); 
	  exit (EXIT_FAILURE);
	  break;
	case -EITI_FMT:
	  fprintf (stderr, "read_procfs: Wrong data format (%s%s).\n",
		   PFS_PREFIX, PFS_STRING); 
	  exit (EXIT_FAILURE);
	  break;
	case -EITI_ENT:
	  fprintf (stderr, "read_procfs: File not found (%s%s).\n",
		   PFS_PREFIX, PFS_STRING);
	  exit (EXIT_FAILURE); 
	  break;
	case -EITI_SYS:
	  fprintf (stderr, "read_procfs: Error while system call (%s).\n",
		   strerror (errno)); 
	  exit (EXIT_FAILURE);
	  break;
	default:
	  fprintf (stderr, "read_procfs: Error not specified!\n"); 
	  exit (EXIT_FAILURE);
	}
    }
  fprintf (stderr, "Driver version: %ld.%ld.%ld-%ld.\n",
	   (version & 0xFF000000) >> 24, (version & 0x00FF0000) >> 16,
	   (version & 0x0000FF00) >> 8, version & 0x000000FF);
  fprintf (stderr, "Compilation Date: %s.\n", date);
  fprintf (stderr, "Boards detected: %d.\n", dev_boards);
  for (board = 0; board < dev_boards; board++)
    {
      switch (dev_modules[board])
	{
	case ICP_AMVS:
	  fprintf (stderr, "Module type board %d: AM-VS.\n", board);
	  break;
	case ICP_AMDG:
	  fprintf (stderr, "Module type board %d: AM-DIG.\n", board);
	  break;
	case ICP_AMPV:
	  fprintf (stderr, "Module type board %d: PCVision.\n", board);
	  break;
	case ITI_VMPCD:
	  fprintf (stderr, "Module type Board %d: PCDig.\n", board);
	  break;
	}
    }

  if (args->board[args->board_i] >= dev_boards)
    {
      fprintf (stderr, "Wrong board selected (%s).\n", optarg);
      exit (EXIT_FAILURE);
    }
  if (*progs > dev_boards)
    {
      fprintf (stderr, "Wrong number of programs selected (%d).\n",
	       *progs);
      exit (EXIT_FAILURE);
    }
  if (args->opmode[args->opmode_i] == MMAP_MODE &&
      args->syncmd[args->syncmd_i] == BLOCK_MODE)
    {
      fprintf (stderr, "If using 'mmap' mode, 'blocking' mode is not");
      fprintf (stderr, " useful. Switching to 'signal' mode.\n");
      args->syncmd[args->syncmd_i] = SIGNAL_MODE;
    }

  if (args->opmode[args->opmode_i] == MMAP_MODE && args->lock[args->lock_i])
    fprintf (stderr, "If using 'mmap' mode, locking is useless.");

  if (args->lock[args->lock_i] && geteuid())
    {
      fprintf (stderr, "One have to be root to lock the read buffer.");
      fprintf (stderr, "Disable locking.");
      args->lock[args->lock_i] = FALSE;
    }
  
  return OK;
}

int
Init_Devs (const struct args_t *args)
{
  int error;
  union iti_dev_t camconf;

  int board, camera, cameras;
  char camera_name[MAX_ENTRY_LENGTH], exo_filename[MAX_ENTRY_LENGTH];

  struct timeval timeout;
  time_t timeout_ms;

  /*
    struct timeval syncrate;
    time_t syncrate_ms;
  */

  char device_name[PATH_MAX];
  
  snprintf (device_name, PATH_MAX, "%s%d%s", ITI_NAME_PREFIX,
	    args->board[args->board_i], ITI_NAME_DATA_POSTFIX);

  if ((devdesc = open (device_name, O_RDONLY)) < 0)
    {
      perror ("Open module device");
      Clean_Up (EXIT_FAILURE);
    }
  
  timeout_ms = 1000 / args->rate[args->rate_i]; 
  MS_TO_TV (timeout_ms, timeout); 
  if (ioctl (devdesc, GIOC_GE_SET_TIMEOUT, &timeout) < 0)
    {
      perror ("GIOC_GE_SET_TIMEOUT");
      Clean_Up (EXIT_FAILURE);
    }

  switch (args->module[args->module_i])
    {
    case ICP_AMVS:
      cameras = VS_CAMERAS_MAX;
      break;
    case ICP_AMDG:
      cameras = DG_CAMERAS_MAX;
      break;
    case ICP_AMPV:
      cameras = PV_CAMERAS_MAX;
      break;
    case ITI_VMPCD:
      cameras = PCD_CAMERAS_MAX;
      break;
    default:
      cameras = 0;
    }

#if (0)
  fprintf (stderr, "\n");
  for (camera = 0; camera < cameras; camera++)
    {
      int wait;

      if (ioctl (devdesc, GIOC_GE_SET_CAMERA, &camera) < 0)
	{
	  if (errno != ENODATA)
	    {
	      perror ("GIOC_GE_SET_CAMERA");
	      Clean_Up (EXIT_FAILURE);
	    }
	}
      for (wait = 0; wait < 3; wait++)
	if (ioctl (devdesc, GIOC_GE_GET_SYNCWAIT, NULL) < 0)
	  {
	    if (errno != ENODATA)
	      {
		perror ("GIOC_GE_GET_SYNCWAIT");
		Clean_Up (EXIT_FAILURE);
	      }
	    
	  }
      if (ioctl (devdesc, GIOC_GE_GET_SYNCWAIT, &syncrate ) < 0)
	{
	  if (errno != ENODATA)
	    {
	      perror ("GIOC_GE_GET_SYNCWAIT");
	      Clean_Up (EXIT_FAILURE);
	    }
	  else
	    {
	      fprintf (stderr, "Camera %d at board %d: Not connected.\n",
		       camera, args->board[args->board_i]);
	    }
	}
      else
	{
	  TV_TO_MS (syncrate, syncrate_ms);
	  fprintf (stderr, "Camera %d at board %d: Syncrate %ldms.\n",
		   camera, args->board[args->board_i], syncrate_ms);
	}
    }
#endif

  if (args->camera[args->camera_i] > cameras)
    {
      fprintf (stderr, "Camera %d not valid on this module type.\n",
	       args->camera[args->camera_i]);
      exit (EXIT_FAILURE);
    }
  if (ioctl (devdesc, GIOC_GE_SET_CAMERA, &args->camera[args->camera_i]) < 0)
    {
      perror ("GIOC_GE_SET_CAMERA");
      Clean_Up (EXIT_FAILURE);
    }

  if (ioctl (devdesc, GIOC_GE_GET_CAMCNF, &camconf) < 0)
    {
      perror ("GIOC_GE_GET_CAMCNF");
      Clean_Up (EXIT_FAILURE);
    }

  if (args->match[args->match_i])
    {
      board = args->board[args->board_i];
      camera = args->camera[args->camera_i];
    }
  else
    {
      board = 0;
      camera = 0;
    }
  
  if ((error = read_config (args->conffile[args->conffile_i], &camconf,
			    board, args->module[args->module_i], camera,
			    camera_name, exo_filename)) < 0)
    {
      switch (error)
	{
	case -EITI_ARG:
	  fprintf (stderr, "read_config: Wrong call Argument\n"
		   "(file: %s, conf: %p, board: %d, module: %s, camera:%d).\n",
		   args->conffile[args->conffile_i], &camconf, board,
		   mod_names[args->module[args->module_i]].name, camera); 
	  exit (EXIT_FAILURE);
	  break;
	case -EITI_FMT:
	  fprintf (stderr, "read_config: Wrong data format (%s).\n",
		   args->conffile[args->conffile_i]); 
	  exit (EXIT_FAILURE);
	  break;
	case -EITI_ENT:
	  fprintf (stderr, "read_config: File not found (%s).\n",
		   args->conffile[args->conffile_i]);
	  exit (EXIT_FAILURE); 
	  break;
	case -EITI_SYS:
	  fprintf (stderr, "read_config: Error while system call (%s).\n",
		   strerror (errno)); 
	  exit (EXIT_FAILURE);
	  break;
	}
      Clean_Up (EXIT_FAILURE);
    }
  fprintf (stderr, "CameraName : %s.\n", camera_name);
  fprintf (stderr, "ExoFilename: %s.\n", exo_filename);
  if (ioctl (devdesc, GIOC_GE_SET_CAMCNF, &camconf) < 0)
    {
      perror ("GIOC_GE_SET_CAMCNF");
      Clean_Up (EXIT_FAILURE);
    }

  switch (args->depth[args->depth_i])
    {
    case 8:
      switch (args->module[args->module_i])
	{
	case ICP_AMDG:
	  if (ioctl (devdesc, GIOC_DG_SET_NORM_TO8, NULL) < 0)
	    {
	      perror ("GIOC_DG_SET_NORM_TO8");
	      Clean_Up (EXIT_FAILURE);
	    }
	  break;
	  
	case ITI_VMPCD:
	  if (ioctl (devdesc, HIOC_PCD_SET_NORM_TO8, NULL) < 0)
	    {
	    perror ("HIOC_PCD_SET_NORM_TO8");
	    Clean_Up (EXIT_FAILURE);
	    }
	  break;
	}
      break;
    case 10:
    case 12:
    case 14:
    case 16:
      switch (args->module[args->module_i])
	{
	case ICP_AMDG:
	  if (ioctl (devdesc, GIOC_DG_SET_LUT_LIN, NULL) < 0)
	    {
	      perror ("GIOC_DG_SET_LUT_LIN");
	      Clean_Up (EXIT_FAILURE);
	    }
	  break;
	case ITI_VMPCD:
	  if (ioctl (devdesc, HIOC_PCD_SET_LUT_LIN, NULL) < 0)
	    {
	      perror ("HIOC_PCD_SET_LUT_LIN");
	      Clean_Up (EXIT_FAILURE);
	    }
	  break;
	case ICP_AMVS:
	case ICP_AMPV:
	  fprintf (stderr, "16bpp not supported at AM-VS and PCVision.\n");
	  Clean_Up (EXIT_FAILURE);
	  break;
	}
      break;
    case 24:
    case 32:
#if (0)
      /* FIXME: this should do the driver! */
      fprintf(stderr, "Setting LUT's !!!\n");
 
      if(ioctl(devdesc, GIOC_GE_GET_CAMCNF, &camconf) < 0)
	{
	  perror ("GIOC_GE_GET_CAMCNF");
	  Clean_Up (EXIT_FAILURE);
	}
      
      camconf.pcd_cam.lut0psz = PCD_PSIZE_16;
      
      if(ioctl(devdesc, GIOC_GE_SET_CAMCNF, &camconf) < 0)
	{
	  perror ("GIOC_GE_SET_CAMCNF");
	  Clean_Up (EXIT_FAILURE);
	}
      
      if(ioctl(devdesc, HIOC_PCD_SET_LUT_LIN, NULL) < 0)
	{
	  perror ("HIOC_PCD_SET_LUT_LIN");
	  Clean_Up (EXIT_FAILURE);
	}
      break;
#endif
    }

  if (ioctl (devdesc, GIOC_CR_SET_HDEC, &args->zoom[args->zoom_i]) < 0)
    {
      perror ("GIOC_CR_SET_HDEC");
      Clean_Up (EXIT_FAILURE);
    }
  if (ioctl (devdesc, GIOC_CR_SET_VDEC, &args->zoom[args->zoom_i]) < 0)
    {
      perror ("GIOC_CR_SET_VDEC");
      Clean_Up (EXIT_FAILURE);
    }

  return OK;
}

int
Setup_Bufs (const struct args_t *args)
{
  if (ioctl (devdesc, GIOC_GE_GET_RAWSIZE, &image.raw_size) < 0)
    {
      perror ("GIOC_GE_GET_RAWSIZE");
      Clean_Up (EXIT_FAILURE);
    }
  if (ioctl (devdesc, GIOC_GE_GET_WIDTH, &image.width) < 0)
    {
      perror ("GIOC_GE_GET_WIDTH");
      Clean_Up (EXIT_FAILURE);
    }
  if (ioctl (devdesc, GIOC_GE_GET_HEIGHT, &image.height) < 0)
    {
      perror ("GIOC_GE_GET_HEIGHT");
      Clean_Up (EXIT_FAILURE);
    }

  fprintf (stderr, "\n");
  fprintf (stderr, "Image size: %d.\n", image.raw_size);
  fprintf (stderr, "Image width: %d.\n", image.width);
  fprintf (stderr, "Image height: %d.\n", image.height);

  if (ioctl (devdesc, GIOC_GE_GET_PAGEDSIZE, &image.paged_size) < 0)
    {
      perror ("GIOC_GE_GET_PAGEDSIZE");
      Clean_Up (EXIT_FAILURE);
    }

  mapsize = args->buffer[args->buffer_i] * image.paged_size; 
  switch (args->opmode[args->opmode_i])
    {
    case READ_MODE:
#if (1)
      if ((zerodesc = open ("/dev/zero", O_RDONLY)) < 0)
	{
	  perror ("Open anonymous mapping device");
	  Clean_Up (EXIT_FAILURE);
	}
      if ((dataptr = (void*)mmap (0, mapsize, PROT_READ | PROT_WRITE,
				  MAP_PRIVATE, zerodesc, 0)) == (void*)-1)
	{
	  perror ("Can't map anonymous device");
	  Clean_Up (EXIT_FAILURE);
	}
#else
      if (!(dataptr = malloc(mapsize)))
	{
	  perror ("Can't allocate save buffer");
	  Clean_Up (EXIT_FAILURE);
	}
#endif
      memset (dataptr, 0x00, mapsize);
      /* if you want some realtime beahivior, locking is useful */
      if (args->lock[args->lock_i])
	if (mlock (dataptr, mapsize) < 0)
	  {
	    perror ("Can't lock anonymous device");
	    Clean_Up (EXIT_FAILURE);
	  }
      break;
    case MMAP_MODE:
      if ((dataptr = (void*)mmap (0, mapsize, PROT_READ,
				  MAP_SHARED, devdesc, 0)) == (void*)-1)
	{
	  perror ("Can't map camera device");
	  Clean_Up (EXIT_FAILURE);
	}
      break;
    }

  image.srcbpp = args->depth[args->depth_i];
  image.srcbpl = image.width * ((image.srcbpp + 7) / 8);
  fprintf (stderr, "Source bits_per_pixel: %d.\n", image.srcbpp);
  fprintf (stderr, "Source bytes_per_line: %d.\n", image.srcbpl);
  
  image.srcptr = dataptr;
  image.srcoff = 0;
  fprintf (stderr, "Source pointer: %p.\n", image.srcptr);
  fprintf (stderr, "Source offset: %ld.\n", image.srcoff);

  return OK;
}

int
Startup_X11 (const struct args_t* args)
{
  int i, j, useshm, screen, planes;
  Colormap colormap;
  XColor color;
  XGCValues values;
  Font timefont;
 
  shminfo.shmaddr = NULL;
 
  if ((display = XOpenDisplay (NULL)) == NULL)
    {
      fprintf (stderr, "Can't open display.\n");
      Clean_Up (EXIT_FAILURE);
    }
   
  screen = DefaultScreen (display);
  planes = DisplayPlanes (display, screen);
  colormap = DefaultColormap (display, screen);
 
 
  if ((window = XCreateSimpleWindow (display, RootWindow (display, screen),
  				     0, 0, image.width, image.height, 4, 
  				     BlackPixel (display, screen),
  				     WhitePixel (display, screen))) == 0)
    {
      fprintf (stderr, "Can't create window.\n");
      Clean_Up (EXIT_FAILURE);
    }	
 
  for (i = 0; i < args->scales[args->scales_i]; i++)
    {
      color.pad = 0;
      color.red = color.green = color.blue =
	(65536 / args->scales[args->scales_i]) * i;
      if (XAllocColor (display, colormap, &color) == 0)
	{
	  fprintf (stderr, "Can't allocate color (%d/%d).\n", i,
		   args->scales[args->scales_i]);
	  Clean_Up (EXIT_FAILURE);
	}
       
      switch (args->scales[args->scales_i])
	{
	case 256:
	  image.lut[i] = color.pixel;
	  break;
	case 128:
	  image.lut[i * 2] = color.pixel;
	  image.lut[i * 2 + 1] = color.pixel;
	  break;
	case 64:
	  for (j = 0; j < 4; j++ )
	    image.lut[i * 4 + j] = color.pixel;
	  break;
	case 32:
	  for (j = 0; j < 8; j++ )
	    image.lut[i * 8 + j] = color.pixel;
	  break;
	case 16:
	  for (j = 0; j < 16; j++ )
	    image.lut[i * 16 + j] = color.pixel;
	  break;
	case 8:
	  for (j = 0; j < 32; j++ )
	    image.lut[i * 32 + j] = color.pixel;
	  break;
	}
    }
   
  if ((context = XCreateGC (display, window, 0, &values)) == NULL)
    {
      fprintf (stderr, "Can't create context.\n");
      Clean_Up (EXIT_FAILURE);
    }
   
  XSetForeground (display, context, WhitePixel(display, screen));
  XSetBackground (display, context, BlackPixel(display, screen));
   
  if ((timefont = XLoadFont (display, FONT_NAME)) == 0)
    {
      fprintf (stderr, "Can't load font.\n");
      Clean_Up (EXIT_FAILURE);
    }
   
  XSetFont (display, context, timefont);
   
  if (args->useshm[args->useshm_i])
    {
      useshm = XShmQueryExtension (display);
      if (!useshm)
	{
	  fprintf (stderr, "XSharedMemoryExtension not available!\n");
	  Clean_Up (EXIT_FAILURE);
	}
       
      ximage = XShmCreateImage (display, DefaultVisual (display, screen),
  				planes, ZPixmap, NULL, &shminfo,
  				image.width, image.height);
       
      shminfo.shmid = shmget (IPC_PRIVATE, 
			      image.raw_size * ximage->bits_per_pixel /
			      (double)image.srcbpp,
  			      IPC_CREAT | 0777);
       
      ximage->data = shminfo.shmaddr = shmat (shminfo.shmid, 0, 0);
      if (!XShmAttach (display, &shminfo))
	{
	  shmdt (shminfo.shmaddr);
	  shmctl (shminfo.shmid, IPC_RMID, 0);
	  fprintf (stderr, "Can't attach shared memory");
	  Clean_Up (EXIT_FAILURE);
	}
    }
  else
    {
      ximage = XCreateImage (display, DefaultVisual (display, screen),
  			     planes, ZPixmap, 0, NULL,
  			     image.width, image.height, 8, image.width);
      if ((ximage->data = malloc (image.raw_size *
				  ximage->bits_per_pixel/(double)image.srcbpp))
	  == NULL)
	{
	  fprintf (stderr, "Virtual memory exhausted.\n");
	  Clean_Up (EXIT_FAILURE);
	}
    }
   
  XMapWindow (display, window);
  XStoreName (display, window, "Live Video");
   
  image.dstbpp = ximage->bits_per_pixel;
  image.dstbpl = ximage->bytes_per_line;
  fprintf (stderr, "Destination bits_per_pixel: %d.\n", image.dstbpp);
  fprintf (stderr, "Destination bytes_per_line: %d.\n", image.dstbpl);
   
  image.dstptr = ximage->data;
  image.dstoff = 0;
   
  fprintf (stderr, "Destination pointer: %p.\n", image.dstptr);
  fprintf (stderr, "Destination offset: %ld.\n", image.dstoff);
   
  XSync (display, TRUE);  
 
  return OK;
}

/* Main */

int
main (int argc, char *argv[])
{
  int i, add, sum, error, boards, create, frame, count;

  struct args_t args =
  {
    0, 0,
    { "unknown.cam" },
    { "unknown" },
    0, 0, 0, 0, 0, 0,
    { FALSE }, { TRUE }, { FALSE }, { TRUE }, { FALSE }, { FALSE },
    0, 0, 0, 0, 0, 0, 0, 0,
    { 0, -1, -1, -1, -1 }, { ICP_AMVS }, { 0 },
    { 8 }, { 64 }, { 1 }, { INT_MAX }, { 1 },
    0,
    { 1.0 },
    0,
    { READ_MODE },
    0,
    { BLOCK_MODE },
    0,
    { NOSPECIAL_MODE }
  };

  int module[ITI_BOARDS_MAX];

  char *basename = argv[0];
  char key;

  struct sigaction copyact, timeact;
  sigset_t local_set;
  struct itimerval wait_itimer;
  fd_set in_fdset, ex_fdset;
  struct timeval wait_time;
  struct pollfd wait_fd;
  int wait_timeout;

  char windowname[NAME_MAX];
  struct timeval thistime, lasttime;
  time_t framediff;
  float framerate, averagesum;
  char ratestring[NAME_MAX];

  caddr_t infoptr = NULL;
  struct iti_info_t *imageinfo;
  struct timeval starttime, timestamp;
  time_t stampdiff;
  char stampstring[9];

  struct iti_acc_t acctinfo;

  volatile struct timeval timediff;	/* else longjmp warning */

  strcpy (args.conffile[0], DEF_CNFDIR);
  for (sum = 0;;)
    {
      add = strcspn (basename, "/");
      if (sum + add++ < strlen (argv[0]))
	sum += add, basename += add;
      else
	break;
    }
  strcat (args.conffile[0], basename);
  strcpy (args.savefile[0], basename);

  i = 0, boards = 1;

  if ((error = Parse_Args (argc, argv, &args, &boards, module)))
    {
      fprintf (stderr, "Error while parsing command line.\n");
      exit (EXIT_FAILURE);
    }

  child = 0;
  children[child] = getpid();
  for (i = 1; i < boards; i++)
    {
      if ((children[i] = fork ()) < 0)
	{
	  perror ("Can't create child process");
	  exit (EXIT_FAILURE);
	}
      if (children[i] == 0)
	{
	  child = i;
	  break;
	}
    }
  
  if (setjmp (jmpstate))
    Clean_Up (EXIT_SIGNAL);

  if (child == 0)
    signal (SIGCHLD, (void*)Sigchld_Catcher);

  signal (SIGINT, (void*)Sigint_Catcher);

  if (args.board_i < child + 1)
    {
      if (args.board_i == 1)
	args.board[child + 1] = args.board[1] + child;
      else
	args.board[child + 1] = child;
    }
  args.board_i = child + 1;
  child++;
  if (args.conffile_i >= child) args.conffile_i = child; 
  if (args.savefile_i >= child) args.savefile_i = child; 
  if (args.save_i >= child) args.save_i = child; 
  if (args.display_i >= child) args.display_i = child; 
  if (args.wait_i >= child) args.wait_i = child; 
  if (args.useshm_i >= child) args.useshm_i = child; 
  if (args.match_i >= child) args.match_i = child; 
  if (args.lock_i >= child) args.lock_i = child; 
  if (args.module_i >= child) args.module_i = child; 
  if (args.camera_i >= child) args.camera_i = child; 
  if (args.depth_i >= child) args.depth_i = child; 
  if (args.scales_i >= child) args.scales_i = child; 
  if (args.buffer_i >= child) args.buffer_i = child; 
  if (args.frames_i >= child) args.frames_i = child; 
  if (args.zoom_i >= child) args.zoom_i = child; 
  if (args.rate_i >= child) args.rate_i = child; 
  if (args.opmode_i >= child) args.opmode_i = child; 
  if (args.syncmd_i >= child) args.syncmd_i = child; 
  if (args.sacqmd_i >= child) args.sacqmd_i = child;
  child--;

  args.module[args.module_i] = module[args.board[args.board_i]];

  fprintf (stderr, "\n");  
  fprintf (stderr, "%d - conffile: %s\n", child,
	   args.conffile[args.conffile_i]);
  fprintf (stderr, "%d - savefile: %s\n", child,
	   args.savefile[args.savefile_i]);
  fprintf (stderr, "%d - save:    %s\n", child,
	   args.save[args.save_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - display: %s\n", child,
	   args.display[args.display_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - wait:    %s\n", child,
	   args.wait[args.wait_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - useshm:  %s\n", child,
	   args.useshm[args.useshm_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - match:  %s\n", child,
	   args.match[args.match_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - lock:  %s\n", child,
	   args.lock[args.lock_i] ? "TRUE" : "FALSE");
  fprintf (stderr, "%d - board:  %d\n", child, args.board[args.board_i]);
  fprintf (stderr, "%d - module: %d\n", child, args.module[args.module_i]);
  fprintf (stderr, "%d - camera: %d\n", child, args.camera[args.camera_i]);
  fprintf (stderr, "%d - depth:  %d\n", child, args.depth[args.depth_i]);
  fprintf (stderr, "%d - scales: %d\n", child, args.scales[args.scales_i]);
  fprintf (stderr, "%d - buffer: %d\n", child, args.buffer[args.buffer_i]);
  fprintf (stderr, "%d - frames: %d\n", child, args.frames[args.frames_i]);
  fprintf (stderr, "%d - zoom:   %d\n", child, args.zoom[args.zoom_i]);
  fprintf (stderr, "%d - rate:   %.2f\n", child, args.rate[args.rate_i]);
  fprintf (stderr, "%d - opmode: %s\n", child,
	  opmode_name[args.opmode[args.opmode_i]].name);
  fprintf (stderr, "%d - syncmd: %s\n", child,
	  syncmd_name[args.syncmd[args.syncmd_i]].name);
  fprintf (stderr, "%d - sacqmd: %s\n", child,
	  sacqmd_name[args.sacqmd[args.sacqmd_i]].name);

  if ((error = Init_Devs (&args)))
    {
      fprintf (stderr, "Error while initializing devices.\n");
      exit (EXIT_FAILURE);
    }
  if ((error = Setup_Bufs (&args)))
    {
      fprintf (stderr, "Error while setup image buffers.\n");
      exit (EXIT_FAILURE);
    }
  
  create = TRUE;
  if (args.save[args.save_i])
    {
      struct stat state;
      
      if (stat (args.savefile[args.save_i], &state) != 0)
	{
	  if (errno != ENOENT)
	    {
	      perror ("Can't get save file state");
	      Clean_Up (EXIT_FAILURE);
	    }
	}
      else
	{
	  if (S_ISBLK(state.st_mode))
	    {
	      if (args.savefile_i < child + 1)
		{
		  fprintf (stderr, "Can't save images concurrently to one"
			   " block device.\n");
		  Clean_Up (EXIT_FAILURE);
		}
	      create = FALSE;
	      if ((rawdesc = open (args.savefile[args.savefile_i],
		  O_WRONLY)) < 0)
		{
		  fprintf (stderr, "Can't open save block device (%s): %s\n",
			   args.savefile[args.savefile_i], strerror (errno));
		  Clean_Up (EXIT_FAILURE);
		}
	    }
	  else
	    {
	      if (!S_ISREG(state.st_mode))
		{
		  fprintf (stderr, "Save file already exists and is not"
			   " a regular file (%s).\n",
			   args.savefile[args.savefile_i]);
		  Clean_Up (EXIT_FAILURE);
		}
	    }
	}
    }

  if (args.display[args.display_i])
    {
      if ((error = Startup_X11 (&args)))
	{
	  fprintf (stderr, "Error while startup X-windows system.\n");
	  exit (EXIT_FAILURE);
	}
    }

  if (args.syncmd[args.syncmd_i] == SIGNAL_MODE)
    {
      ignore.sa_handler = SIG_IGN;
      sigemptyset (&ignore.sa_mask);
      ignore.sa_flags = 0;
      if (sigaction (SIGIO, &ignore, NULL) < 0)
	{
	  perror ("Can't install IGNORE handler");
	  Clean_Up (EXIT_FAILURE);
	}
      copyact.sa_handler = (void*)Sigio_Catcher;
      sigemptyset (&copyact.sa_mask);
#ifndef __NetBSD__
      copyact.sa_flags = SA_ONESHOT;
#endif
      timeact.sa_handler = (void*)Sigalrm_Catcher;
      sigemptyset (&timeact.sa_mask);
#ifndef __NetBSD__
      timeact.sa_flags = SA_ONESHOT;
#endif
      if (fcntl (devdesc, F_SETOWN, getpid()) < 0)
	{
	  perror ("F_SETOWN");
	  Clean_Up (EXIT_FAILURE);
	}
      if (fcntl (devdesc, F_SETFL,
		 fcntl (devdesc, F_GETFL, NULL) | FASYNC) < 0)
	{
	  perror ("F_SETFL FASYNC");
	  Clean_Up (EXIT_FAILURE);
	}
    }

  if (args.syncmd[args.syncmd_i] != BLOCK_MODE)
    {
      if (fcntl (devdesc, F_SETFL, 
		 fcntl (devdesc, F_GETFL, NULL) | O_NONBLOCK) < 0)
	{
	  perror ("F_SETFL O_NONBLOCK");
	  Clean_Up (EXIT_FAILURE);
	}
    }


  if (args.opmode[args.opmode_i] == MMAP_MODE)
    {
      if (ioctl (devdesc, GIOC_CR_SET_START, NULL) < 0)
	{
	  perror ("GIOC_CR_SET_START");
	  Clean_Up (EXIT_FAILURE);
	}
    }
  
  snprintf (stampstring, 9, "00:00:00");
  gettimeofday (&starttime, NULL);
  
  if (args.save[args.save_i] && !args.wait[args.wait_i])
    {
      signal (SIGQUIT, (void*)Sigquit_Catcher);
      fprintf (stderr, "Live mode activated.\n");
      fprintf (stderr, "Give SIGQUIT (usually Ctrl+\\)"
	       " to start image recording.\n");
    }

  frame = 0;
  averagesum = 0.0;
  while (frame < args.frames[args.frames_i])
    {

      /* image acquisition */
      timeout = FALSE;

      if  (args.opmode[args.opmode_i] == READ_MODE)
	{
	  if (read (devdesc, image.srcptr,
		    args.buffer[args.buffer_i] * image.paged_size) < 0)
	    {
	      if (errno == ENODATA)
		timeout = TRUE;
	      else
		{
		  if (errno == EINTR)
		    goto done;
		  else
		    {
		      perror ("Can't read image");
		      Clean_Up (EXIT_FAILURE);
		    }
		}
	    }
	  else
	    ITI_PRINTF("-");
	}
      
      switch (args.syncmd[args.syncmd_i])
	{
	case BLOCK_MODE:
	  break;
	case SIGNAL_MODE:
	  if (sigaction (SIGIO, &copyact, NULL) < 0)
	    {
	      perror ("Can't install SIGIO handler");
	      Clean_Up (EXIT_FAILURE);
	    }
	  if (sigaction (SIGALRM, &timeact, NULL) < 0)
	    {
	      perror ("Can't install SIGALRM handler");
	      Clean_Up (EXIT_FAILURE);
	    }
	  wait_itimer.it_interval.tv_sec = 0;
	  wait_itimer.it_interval.tv_usec = 0;
	  wait_itimer.it_value.tv_sec =
	    floor (3 / args.rate[args.rate_i]);
	  wait_itimer.it_value.tv_usec =
	    fmod (3 / args.rate[args.rate_i], 1.0) * 1000000;
	  if (setitimer (ITIMER_REAL, &wait_itimer, NULL) != 0)
	    {
	      perror ("Can't install intervall timer");
	      Clean_Up (EXIT_FAILURE);
	    }
	  sigemptyset (&local_set);
	  if (sigsuspend (&local_set) < 0)
	    {
	      if (errno != EINTR)
		{
		  perror ("Can't call sigsuspend");
		  Clean_Up (EXIT_FAILURE);
		}
	    }
	  if (sigaction (SIGIO, &ignore, NULL) < 0)
	    {
	      perror ("Can't install IGNORE handler");
	      Clean_Up (EXIT_FAILURE);
	    }
	  if (!timeout)
	    {
	      ITI_PRINTF("\\");
	      wait_itimer.it_value.tv_sec = 0;
	      wait_itimer.it_value.tv_usec = 0;
	      if (setitimer (ITIMER_REAL, &wait_itimer, NULL) != 0)
		{
		  perror ("Can't remove intervall timer");
		  Clean_Up (EXIT_FAILURE);
		}
	    }
	  break;
	case SELECT_MODE:
	  FD_ZERO (&in_fdset);
	  FD_ZERO (&ex_fdset);
	  FD_SET (devdesc, &in_fdset);
	  FD_SET (devdesc, &ex_fdset);
	  wait_time.tv_sec =
	    floor (3 / args.rate[args.rate_i]);
	  wait_time.tv_usec =
	    fmod (3 / args.rate[args.rate_i], 1.0) * 1000000;
	  error = select (getdtablesize (), &in_fdset, NULL, &ex_fdset,
			  &wait_time);
	  switch (error)
	    {
	    case -1:
	      perror ("Some error during select occured");
	      Clean_Up (EXIT_FAILURE);
	      break;
	    case 0:
	      timeout = TRUE;
	      break;
	    case 1:
	    case 2:
	      if (FD_ISSET (devdesc, &in_fdset))
		ITI_PRINTF("/");
	      if (FD_ISSET (devdesc, &ex_fdset))
		timeout = TRUE;
	      break;
	    default:
	      perror ("unexpected select return value");
	      Clean_Up (EXIT_FAILURE);
	    }
	  break;
	case POLL_MODE:
	  wait_fd.fd = devdesc;
	  wait_fd.events = POLLIN | POLLPRI;
	  wait_timeout = (1 / args.rate[args.rate_i]) * 1000;
	  error = poll (&wait_fd, 1, wait_timeout);
	  switch (error)
	    {
	    case -1:
	      perror ("Some error during poll occured");
	      Clean_Up (EXIT_FAILURE);
	      break;
	    case 0:
	      timeout = TRUE;
	      break;
	    case 1:
	      if (wait_fd.revents & POLLIN)
		ITI_PRINTF("/");
	      if (wait_fd.revents & POLLPRI)
		timeout = TRUE;
	      break;
	    default:
	      perror ("unexpected poll return value");
	      Clean_Up (EXIT_FAILURE);
	    }
	  break;
	}

      if (args.opmode[args.opmode_i] == READ_MODE &&
	  args.syncmd[args.syncmd_i] != BLOCK_MODE)
	{
	  if (read (devdesc, image.srcptr,
		    args.buffer[args.buffer_i] * image.paged_size) < 0)
	    {
	      if (errno == ENODATA)
		timeout = TRUE;
	      else
		{
		  if (errno == EINTR)
		    goto done;
		  else
		    {
		      perror ("Can't read image");
		      Clean_Up (EXIT_FAILURE);
		    }
		}
	    }
	  else
		ITI_PRINTF("-");
	}

      /* image processing */
      count = 0;
      do 
  	{
	  switch (args.opmode[args.opmode_i])
	    {
	    case READ_MODE:
	      image.srcoff = count * image.paged_size;
	      break;
	    case MMAP_MODE:
	      do
		if ((image.srcoff = lseek (devdesc, 0UL, SEEK_CUR)) < 0)
		  {
		    if (errno != EINTR)
		      {
			perror ("lseek 0/SEEK_CUR");
			Clean_Up (EXIT_FAILURE);
		      }
		  }
		else
		  break;
	      while (TRUE);
	      break;
	    }
	  ITI_PRINTF ("%ld", image.srcoff / image.paged_size);

	  if (args.display[args.display_i])
	    {
	      if (Display_Image (&args) < 0)
		Clean_Up (EXIT_FAILURE);
	  
	      if (args.useshm[args.useshm_i])
		XShmPutImage (display, window, context, ximage,
			      0, 0, 0, 0, image.width, image.height, 0);
	      else
		XPutImage (display, window, context, ximage,
			   0, 0, 0, 0, image.width, image.height);
	      if (timeout)
		{
		  strcpy (windowname, "*** No Sync ***");
		  XStoreName (display, window, windowname);
		  snprintf (stampstring, 9, "--:--:--");
		  XDrawImageString (display, window, context,
				    40, 30, stampstring, 8);
		}
	      else
		{
		  if (frame)
		    {
		      gettimeofday (&thistime, NULL);
		      TV_MINUS(timediff, thistime, lasttime);
		      TV_TO_MS (timediff, framediff);
		      memcpy (&lasttime, &thistime, sizeof(struct timeval));
		      framerate = 1000.0 / (float)framediff;
		      averagesum += framerate;
		      strcpy (windowname, "Frames/sec: ");
		      snprintf (ratestring, NAME_MAX, "%3.1f (%3.1f)",
				framerate, averagesum / frame);
		      strcat (windowname, ratestring);
		      XStoreName (display, window, windowname);
		    }
		  infoptr = image.srcptr + image.srcoff;
		  imageinfo = (struct iti_info_t*)
		    (infoptr + image.srcbpl * image.height);
		  memcpy (&timestamp, &(imageinfo->timestamp),
			  sizeof(struct timeval));
		  TV_MINUS (timediff, timestamp, starttime);
		  TV_TO_MS (timediff, stampdiff);
		  snprintf (stampstring, 9, "%2ld:%2ld:%2ld",
			   (stampdiff / 60000) % 60,
			   (stampdiff / 1000) % 60,
			   (stampdiff / 10) % 100);
		  XDrawImageString (display, window, context,
				    40, 30, stampstring, 8);
		}
	      XSync (display, FALSE);
	    }
      	  
	  if ((args.save[args.save_i] && !args.display[args.display_i]) ||
	      (args.save[args.save_i] && !args.wait[args.wait_i] && record))
	    if (Save_Image (&args, frame, count, stampstring, create) < 0)
	      Clean_Up (EXIT_FAILURE);
      
	  if (args.wait[args.wait_i])
	    {
	      if (args.display[args.display_i] && !args.save[args.save_i])
		{
		  fprintf (stderr, "Display next image?"
			   " (CR for yes / 'q'+CR for quit)\n");
		  while ((key = getchar ()) == 0);
		  if (key == 'q') goto done;
		}
	      else
		if (args.save[args.save_i] && !args.display[args.display_i])
		  {
		    fprintf (stderr, "Save next image?"
			     " (CR for yes / 'q'+CR for quit)\n");
		    while ((key = getchar ()) == 0);
		    if (key == 'q') goto done;
		  }
		else
		  {
		    fprintf (stderr, "Save this and display next image?\n");
		    fprintf (stderr, "(CR for ONLY display /"
			     " `y'+CR for ALSO save / 'q'+CR for quit)\n");
		    while ((key = getchar ()) == 0);
		    if (key == 'q') goto done;
		    if (key == 'y')
		      {
			if (Save_Image (&args, frame, count,
					stampstring, create) < 0)
			  Clean_Up (EXIT_FAILURE);
			key = getchar ();
		      }
		  }
	    }

	  count++;
	}
      while (args.opmode[args.opmode_i] == READ_MODE &&
	     count < args.buffer[args.buffer_i]);
      
      image.srcoff = 0;
      frame++;
    }

 done:

  if (args.opmode[args.opmode_i] == MMAP_MODE)
    {
      if (ioctl (devdesc, GIOC_CR_SET_STOP, NULL) < 0)
	{
	  perror ("GIOC_CR_SET_STOP");
	  Clean_Up (EXIT_FAILURE);
	}
    }
  
  if (ioctl (devdesc, GIOC_GE_GET_STATS, &acctinfo) < 0)
    {
      perror ("GIOC_GE_GET_STATS");
      exit (EXIT_FAILURE);
    }
  fprintf (stderr, "\nrun statistics:\n");
  fprintf (stderr, "\tframes captured %ld,\n",
	   acctinfo.frames_captured);
  fprintf (stderr, "\tframes transfered %ld,\n",
	   acctinfo.frames_transfered);
  fprintf (stderr, "\tframes timedout %ld,\n",
	   acctinfo.frames_timedout);
  fprintf (stderr, "\tframes get_by_blocked %ld.\n",
	   acctinfo.frames_get_by_blocked);
  fprintf (stderr, "\tframes get_by_signal %ld.\n",
	   acctinfo.frames_get_by_signal);
  fprintf (stderr, "\tframes get_by_select %ld.\n",
	   acctinfo.frames_get_by_select);
  
  return (EXIT_SUCCESS);
}

